import optparse
import subprocess
import sys
import unittest

import django
import subunit
import testtools
from django.test import runner
from ucitests import (
    filters,
    results,
    runners,
    suites,
)


with_argparse = django.VERSION[0:2] >= (1, 8)


def validate_pick(option, opt_str, value, parser, *args, **kwargs):

    def BadValue(msg):
        msg_with_opt = '{}: {}'.format(str(option), msg)
        return optparse.OptionValueError(msg_with_opt)

    try:
        rank, modulo = value.split('/')
    except ValueError:
        raise BadValue('{} does not match n/m format'.format(value))

    def check_int(val):
        try:
            return abs(int(val))
        except ValueError:
            raise BadValue('{} is not an integer'.format(val))

    rank, modulo = [check_int(rank), check_int(modulo)]
    if not rank or not modulo:
        raise BadValue('Zero makes no sense in {}'.format(value))
    if rank > modulo:
        raise BadValue('{} must be <= {} in {}'.format(rank, modulo, value))
    setattr(parser.values, option.dest, [rank, modulo])


class TestRunner(runner.DiscoverRunner):

    wrapping_test_suite = unittest.TestSuite
    split_suites = dict(fork=suites.ForkedSuite)

    if not with_argparse:
        option_list = runner.DiscoverRunner.option_list + (
            optparse.make_option(
                '-c', '--concurrency',
                action='store', dest='concurrency', type='int',
                default=1,
                help='How many concurrent processes to run the tests.'),
            optparse.make_option(
                '-l', '--list',
                action='store_true', dest='list_only',
                default=False,
                help='List test ids without running them.'),
            optparse.make_option(
                '-i', '--include',
                action='append', dest='include_regexps',
                default=None,
                help='Tests matching this regexp will be run.'),
            optparse.make_option(
                '-X', '--exclude',
                action='append', dest='exclude_regexps',
                default=None,
                help='Tests matching this regexp will not be run.'),
            optparse.make_option(
                '--load',
                action='store', dest='load_file',
                default=None,
                help='A file name containing test ids to be run.'),
            optparse.make_option(
                '-f', '--format',
                action='store', dest='format', type='choice',
                choices=['text', 'subunit'], default='text',
                help='Output format for the test results.'),
            optparse.make_option(
                '-s', '--sub-runner',
                action='store', dest='sub_runner', default='fork',
                help='The concurrent test runner to use.'),
            optparse.make_option(
                '--pick',
                dest='pick', default=None, type="string",
                action='callback', callback=validate_pick,
                help='n/m pick nth test every m tests.'),
        )

    def __init__(self, pattern=None, top_level=None, verbosity=1,
                 interactive=True, failfast=False, concurrency=1,
                 list_only=False, include_regexps=None, exclude_regexps=None,
                 load_file=None, format=None,
                 sub_runner=None, pick=None,
                 output=None,
                 **kwargs):
        super(TestRunner, self).__init__(
            pattern, top_level, verbosity=verbosity, interactive=interactive,
            failfast=failfast, **kwargs)
        self.list_only = list_only
        if output is None:
            output = sys.stdout
        self.output = output
        self.include_regexps = include_regexps
        self.exclude_regexps = exclude_regexps
        self.concurrency = concurrency
        self.load_file = load_file
        self.format = format
        self.sub_runner = sub_runner
        self.pick = pick

    @classmethod
    def add_arguments(cls, parser):
        if with_argparse:
            runner.DiscoverRunner.add_arguments(parser)
            runners.add_uci_run_tests_options(parser)
            parser.add_argument(
                '--sub-runner', '-s',
                action='store', dest='sub_runner', default='fork',
                help='The concurrent test runner to use.')

    def setup_databases(self, **kwargs):
        """Override base class to defer the db setup.

        The db setup id done in SSOTestSuite so each worker can use its own
        db.

        """
        # Return an empty but valid value so teardown_databases does nothing
        # when invoked from the base class run_suite().
        return ([], [])

    def build_suite(self, test_labels=None, extra_tests=None, **kwargs):
        if not test_labels:
            # Without this, only the tests in the acceptance folder
            # will be tried to be run when no labels are specified.
            test_labels = ('src',)
        suite = super(TestRunner, self).build_suite(
            test_labels, extra_tests, **kwargs)
        suite = filters.include_regexps(self.include_regexps, suite)
        suite = filters.exclude_regexps(self.exclude_regexps, suite)
        if self.load_file is not None:
            # Load the file and filter the suite
            with open(self.load_file, 'r') as f:
                id_list = filters.TestIdList(filters.load_test_id_list(f))
                suite = filters.include_id_list(id_list, suite)
        if self.pick is not None:
            rank, modulo = self.pick
            suite = filters.pick(suite, rank, modulo)
        return self.wrapping_test_suite(suite)

    def run_tests(self, test_labels, extra_tests=None, **kwargs):
        # Override base class so we can list the tests without the costly
        # setup.
        if self.list_only:
            suite = self.build_suite(test_labels, extra_tests)
            for t in filters.iter_flat(suite):
                self.output.write(t.id() + '\n')
            return 0
        # Otherwise let the base class handle the setup and the run
        return super(TestRunner, self).run_tests(test_labels, extra_tests=None,
                                                 **kwargs)

    def run_suite(self, suite, **kwargs):
        initial_test_nb = suite.countTestCases()
        if self.format == 'text':
            result = results.TextResult(sys.stdout,
                                        verbosity=self.verbosity,
                                        failfast=self.failfast)
        else:
            result = testtools.ExtendedToStreamDecorator(
                subunit.StreamResultToBytes(sys.stdout))
        suite.addCleanup(
            self.check_all_tests_executed, initial_test_nb, result)
        runners.run_tests(suite, result, self.concurrency,
                          self.split_suites[self.sub_runner])
        return result

    def check_all_tests_executed(self, initial_test_nb, result):
        # In very rare failure modes, tests that should be run may not
        # increment 'testsRun'. This sanity check will catch them.
        try:
            if initial_test_nb != result.testsRun:
                raise AssertionError(
                    'Expecting {}, got {} tests run'.format(
                        initial_test_nb, result.testsRun))
        except AssertionError:
            fake_id = '{}.{}.check_all_tests_executed'.format(
                self.__module__, self.__class__.__name__)
            err = testtools.ErrorHolder(fake_id, sys.exc_info())
            err.run(result)
