import os
import tempfile
import unittest
from cStringIO import StringIO


from django.core import management
from ucitests import (
    filters,
    fixtures,
)

from olstestsdjango import runner


class FakeRunner(runner.TestRunner):
    """Fake test runner for tests.

    This ensures we don't accidentally run tests while involving the django
    commands framework while allowing tests to check how the runner is created.
    """
    # So tests can capture runner creations
    init_hook = None

    def __init__(self, *args, **kwargs):
        super(FakeRunner, self).__init__(*args, **kwargs)
        if self.init_hook is not None:
            self.init_hook(self, *args, **kwargs)

    def run_tests(self, *args, **kwargs):
        # Overriding the base class ensures we don't call the setup which is
        # both costly and may have side-effects.
        return 0


class TestRunnerWithFake(unittest.TestCase):

    def setUp(self):
        super(TestRunnerWithFake, self).setUp()

        def capture(runner, *args, **kwargs):
            # We only care about the runner
            self.runner = runner

        fixtures.patch(self, FakeRunner, 'init_hook', capture)

    def run_test(self, args):
        app_name = management.get_commands()['test']
        test_cmd = management.load_command_class(app_name, 'test')
        test_runner_name = FakeRunner.__module__ + '.FakeRunner'
        test_cmd.run_from_argv(
            # boilerplate
            ['manage.py', 'test',
             # force our fake runner
             '--testrunner=%s' % (test_runner_name,),
             # restrict to our own app
             'olstestsdjango']
            # Whatever we want to test
            + args)
        return test_cmd

    def test_defaults(self):
        self.run_test([''])
        self.assertEqual(1, self.runner.concurrency)
        self.assertEqual(None, self.runner.include_regexps)
        self.assertEqual(None, self.runner.exclude_regexps)
        self.assertEqual('text', self.runner.format)
        self.assertEqual(False, self.runner.list_only)
        self.assertIs(None, self.runner.load_file)
        self.assertEqual('fork', self.runner.sub_runner)
        self.assertIs(None, self.runner.pick)

    def test_list(self):
        self.run_test(['--list'])
        self.assertEqual(True, self.runner.list_only)

    def test_include(self):
        my_id = self.id()
        self.run_test(['--include', my_id, '-i', 'foo'])
        self.assertEqual([my_id, 'foo'], self.runner.include_regexps)

    def test_exclude(self):
        my_id = self.id()
        self.run_test(['--exclude', my_id, '-X', 'foo'])
        self.assertEqual([my_id, 'foo'], self.runner.exclude_regexps)

    def test_concurrency_short(self):
        self.run_test(['-c8'])
        self.assertEqual(8, self.runner.concurrency)

    def test_concurrency_long(self):
        self.run_test(['--concurrency', '4'])
        self.assertEqual(4, self.runner.concurrency)

    @unittest.skipIf(runner.with_argparse, 'Pointless for argparse')
    def test_load_list(self):
        self.run_test(['--load', 'test.list'])
        self.assertEqual('test.list', self.runner.load_file)

    def test_format_subunit(self):
        self.run_test(['--format', 'subunit'])
        self.assertEqual('subunit', self.runner.format)

    def test_sub_runner_lxc(self):
        self.run_test(['--sub-runner', 'lxc'])
        self.assertEqual('lxc', self.runner.sub_runner)

    def test_pick(self):
        self.run_test(['--pick', '1/2'])
        self.assertEqual([1, 2], self.runner.pick)


class TestBuildSuite(unittest.TestCase):

    def get_test_names(self, labels=None, **kwargs):
        if labels is None:
            labels = ['olstestsdjango']
        out = StringIO()
        test_runner = runner.TestRunner(list_only=True, output=out, **kwargs)
        # We don't need more that our own app test suite
        suite = test_runner.build_suite(labels)
        return [t.id() for t in filters.iter_flat(suite)]

    def test_list_one_test(self):
        my_id = self.id()
        self.assertEqual([my_id], self.get_test_names(include_regexps=[my_id]))

    def test_list_module(self):
        my_module_name = self.__class__.__module__
        names = self.get_test_names(include=my_module_name)
        # Checking all module tests here will make the test brittle, settling
        # for a single one and making sure there are more.
        self.assertTrue(self.id() in names)
        self.assertLessEqual(2, len(names))
        # All the tests should be in our module
        for name in names:
            self.assertTrue(name.startswith(my_module_name))

    def test_load_unknown_file(self):
        with self.assertRaises(IOError) as cm:
            self.get_test_names(load_file='I-dont-exist')
        self.assertEqual('I-dont-exist', cm.exception.filename)

    def test_load_file(self):
        _, filename = tempfile.mkstemp(prefix='my-test-list.')
        self.addCleanup(os.remove, filename)
        my_id = self.id()
        with open(filename, 'w') as f:
            f.write('{}\n'.format(my_id))
        self.assertEqual([my_id], self.get_test_names(load_file=filename))


class TestRealRunner(unittest.TestCase):
    """Integration tests for TestRunner.

    Since we don't want to run the test run setup (costly), we settle on making
    sure the right tests are loaded and listed.
    """

    def test_list_one_test(self):
        my_id = self.id()
        out = StringIO()
        test_runner = runner.TestRunner(list_only=True, output=out,
                                        include_regexps=[my_id])
        # We don't need more that our own app test suite
        test_runner.run_tests(['olstestsdjango'])
        self.assertEqual([my_id], out.getvalue().splitlines())
